include Math

STOP = 0xff
GET_PLANE = 0
TEXTURE = 1
RENDER = 2
TESSREPEAT = 3
MOVEX = 4
MOVEY = 5
MOVEZ = 6
CALL = 7
RETURN = 8
LOOP = 9
ENDLOOP = 10
ROTX = 11
ROTY = 12
ROTZ = 13
SCALEX = 14
SCALEY = 15
SCALEZ = 16
SCALE = 17
SCALEU = 18
SCALEV = 19
SCALEUV = 20
TUBE = 21
CUBE = 22
REFLECT = 23
DISC = 24
BLENDFUNC = 25
XOFFSET = 26
YOFFSET = 27
SETFRAME = 28
BLITFROM = 29
SOLID = 30
INIT = 31

class Stream
	attr_accessor :flags

	def initialize
		@stream = ""
		@flags = 0
	end

	def add(i, f = 0)
		@stream += (i + @flags).chr + [f].pack("f")[1, 3]
	end

	# align and return qword position
	def pos
		add(0) while (@stream.size & 15) != 0
		@stream.size / 16
	end

	def write(filename)
		pos
		File.open(filename, "wb") {|f| f.write(@stream)}
	end
end

# read texturesymbols
textures = {}
File.foreach("../texgen/texturesymbols.txt") do |line|
	name, buffer = line.split
	textures[name] = buffer.to_i
end

# read file
file = ""
File.open(ARGV[0], "r") {|f| file = f.read}

labels = {}
stream = nil

2.times do |pass|
	stream = Stream.new
	file.each do |line|
		line.sub!(/;.*/, "")
		next if line.strip == ""
		m, p = line.split(' ', 2)
  		p.strip!
		match = /time\((.*)\)/.match(p)
		if match && match.begin(0) == 0
			stream.flags = 128
			p = match[1]
		else
			stream.flags = 0
		end
		if m[-1, 1] == ":"
			labels[m[0, m.size-1]] = stream.pos
		else
			case m
				when "end"
					stream.add TESSREPEAT
					stream.add STOP
				when "plane"
					stream.add TESSREPEAT
					stream.add GET_PLANE, p.to_i
				when "texture"
					raise "Unknown texture '#{p}'" unless textures[p]
					stream.add TEXTURE, textures[p]
				when "render"
					stream.add RENDER
				when "movex"
					stream.add MOVEX, eval(p)
				when "movey"
					stream.add MOVEY, eval(p)
				when "movez"
					stream.add MOVEZ, eval(p)
				when "call"
					if labels[p]
						stream.add CALL, labels[p]
					else
						raise "Unknown label '#{p}'" if pass != 0
						stream.add CALL, 0
					end
				when "return"
					stream.add RETURN
				when "loop"
					stream.add LOOP, p.to_i
				when "endloop"
					stream.add ENDLOOP
				when "rotx"
					stream.add ROTX, eval(p)
				when "roty"
					stream.add ROTY, eval(p)
				when "rotz"
					stream.add ROTZ, eval(p)
				when "scale"
					stream.add SCALE, eval(p)
				when "scalex"
					stream.add SCALEX, eval(p)
				when "scaley"
					stream.add SCALEY, eval(p)
				when "scalez"
					stream.add SCALEZ, eval(p)
				when "scaleu"
					stream.add SCALEU, eval(p)
				when "scalev"
					stream.add SCALEV, eval(p)
				when "scaleuv"
					stream.add SCALEUV, eval(p)
				when "tube"
					stream.add TESSREPEAT
					stream.add TUBE, p.to_i
				when "cube"
					stream.add TESSREPEAT
					stream.add CUBE, eval(p)
				when "reflect"
					stream.add REFLECT
				when "disc"
					stream.add TESSREPEAT
					stream.add DISC, p.to_i
				when "tessrepeat"
					stream.add TESSREPEAT
				when "blendfunc"
					stream.add BLENDFUNC, p.to_i
				when "xoffset"
					stream.add XOFFSET, eval(p)+0.5
				when "yoffset"
					stream.add YOFFSET, eval(p)+0.5
				when "setframe"
					stream.add SETFRAME, p.to_i
				when "blitfrom"
					stream.add BLITFROM, p.to_i
				when "solid"
					stream.add SOLID, eval(p)
				when "init"
					stream.add INIT
				else
					raise "Unknown instruction '#{m}'"
			end
		end
	end
end

stream.write(ARGV[1])

File.open("geolabels.txt", "w") do |f|
	labels.each {|l, a| f.puts("#{l} #{a}")}
end
