.syntax new

SELECT		.equ	H'1
START		.equ	H'8
UP			.equ	H'10
RIGHT		.equ	H'20
DOWN		.equ	H'40
LEFT		.equ	H'80
L2			.equ	H'100
R2			.equ	H'200
L1			.equ	H'400
R1			.equ	H'800
TRIANGLE	.equ	H'1000
CIRCLE		.equ	H'2000
CROSS		.equ	H'4000
SQUARE		.equ	H'8000

.macro	testpad key, label
.aif	\key EQ H'8000
	iaddiu		mask, vi0, 0x4000
	iadd		mask, mask, mask
.aelse
	iaddiu		mask, vi0, \key
.aendi
	iand		tmp, mask, pad
	ibne		tmp, vi0, \label
.endm

.macro fakesin out, in		; in [0..1] (full circle), out [-1..1]
	loi			4
	mul			acc, \in, i
	loi			0x800000*4
	madd		acc, one, i
	msub		acc, one, i
	loi			4
	msub		acc, \in, i
	loi			3
	madd		sintmp, one, i
	max			sintmp2, sintmp, one
	mini		sintmp, sintmp, one
	loi			2
	sub			sintmp2, sintmp2, i
	mul			sintmp2, sintmp2, sintmp2
	mul			acc, sintmp, sintmp
	msub		\out, one, sintmp2
.endm

.macro saveposition
	sq			scriptdata, loopstack(vi0)
	isw.x		script, loopstack+1(vi0)
	isw.y		currentword, loopstack+1(vi0)
.endm

.macro pushloop
	sq			scriptdata, 0(loopsp)
	isw.x		script, 1(loopsp)
	isw.y		currentword, 1(loopsp)
	iaddiu		loopsp, loopsp, 2
.endm

.macro poploop
	isubiu		loopsp, loopsp, 2
	lq			scriptdata, 0(loopsp)
	ilw.x		script, 1(loopsp)
	ilw.y		currentword, 1(loopsp)
.endm

.macro normalize reg
	mul.yz		dot, \reg, \reg
	add.x		one, vf0, vf0[w]
	mul.x		acc, \reg, \reg
	madd.x		acc, one, dot[y]
	madd.x		dot, one, dot[z]
	rsqrt		q, vf0[w], dot[x]
	mul.xyz		\reg, \reg, q
.endm

.include "vcl_sml.inc"

--enter
	.init_vi_all
	.init_vf_all
--endenter

	fcset		0

	; inc time
	lq.xy		time, timeval(vi0)
	loi			1.0/60.0
	add.xy		time, time, i
	sq.xy		time, timeval(vi0)

; *** pad code
	ilw.w		cnt, 0x3ff(vi0)
	isubiu		mone, vi0, 1
	isub		cnt, mone, cnt
	ilw.x		mask, keys(vi0)
	isub		mask, mone, mask
	iand		trg, cnt, mask
	isw.x		cnt, keys(vi0)
	isw.y		trg, keys(vi0)
	
	iaddiu		tmp, vi0, START
	iand		tmp, trg, tmp
	ibeq		tmp, vi0, nostart
	iaddiu		tmp, vi0, 1
	isw.w		tmp, keys(vi0)
nostart:

; **** move camera

; test if we are in auto mode (default)
	ilw.w		tmp, keys(vi0)
	ibeq		tmp, vi0, camauto

	lq			cammat[3], cameramatrix+3(vi0)
	lq			viewangle, cameramatrix+4(vi0)
	ilw.w		pad, 0x3ff(vi0)
	; ** rotate
	testpad		CIRCLE, cnotcircle
	loi			0.02
	add.x		viewangle, viewangle, i
cnotcircle:
	testpad		SQUARE, cnotsquare
	loi			-0.02
	add.x		viewangle, viewangle, i
cnotsquare:
	testpad		TRIANGLE, cnottriangle
	loi			-0.02
	add.z		viewangle, viewangle, i
cnottriangle:
	testpad		CROSS, cnotcross
	loi			0.02
	add.z		viewangle, viewangle, i
cnotcross:
	sq			viewangle, cameramatrix+4(vi0)
	; ** build the transformation part of the matrix from euler angles
	AngleSinCos	viewangle, sincos
	mfir		cammat[0], vi0
	mfir		cammat[1], vi0
	mfir		cammat[2], vi0
	sub.y		cammat[1], vf0, sincos[w]
	sub.z		cammat[1], vf0, sincos[z]
	sub.y		cammat[2], vf0, sincos[z]
	add.z		cammat[2], vf0, sincos[w]
	add.x		cammat[0], vf0, vf0[w]
	add.x		rotvecx, vf0, sincos[y]
	sub.z		rotvecx, vf0, sincos[x]
	add.x		rotvecz, vf0, sincos[x]
	add.z		rotvecz, vf0, sincos[y]
	mul.xz		acc, rotvecx, cammat[0][x]
	madd.xz		cammat[0], rotvecz, cammat[0][z]
	mul.xz		acc, rotvecx, cammat[1][x]
	madd.xz		cammat[1], rotvecz, cammat[1][z]
	mul.xz		acc, rotvecx, cammat[2][x]
	madd.xz		cammat[2], rotvecz, cammat[2][z]
	; ** move in all 3 axes
	mul			acc, cammat[3], vf0[w]
	testpad		UP, cnotup
	loi			0.02
	madd		acc, cammat[2], i
cnotup:
	testpad		DOWN, cnotdown
	loi			-0.02
	madd		acc, cammat[2], i
cnotdown:
	testpad		LEFT, cnotleft
	loi			-0.02
	madd		acc, cammat[0], i
cnotleft:
	testpad		RIGHT, cnotright
	loi			0.02
	madd		acc, cammat[0], i
cnotright:
	testpad		R1, cnotr1
	loi			-0.02
	madd		acc, cammat[1], i
cnotr1:
	testpad		R2, cnotr2
	loi			0.02
	madd		acc, cammat[1], i
cnotr2:
	madd		cammat[3], vf0, vf0[x]
	b			camskipauto

camauto:

	ilw.x		ptr, splineptr(vi0)
	lq.x		t, timeval(vi0)
	iaddiu		mask, vi0, CROSS
	iand		tmp, mask, trg
	ibne		tmp, vi0, newspline
	ftoi0.x		tmp, t
	mtir		tmp, tmp[x]
	isubiu		tmp, tmp, 3
	ibltz		tmp, keepspline
newspline:
	sub.x		t, vf0, vf0
	sq.x		t, timeval(vi0)
	iaddiu		ptr, ptr, 12
	ilw.w		tmp, 0(ptr)
	iaddiu		tmp, tmp, 1
	ibne		tmp, vi0, keepptr
	iaddiu		ptr, vi0, splines
keepptr:
	ilw.w		tmp, 2(ptr)
	iaddiu		tmp, tmp,1
	ibeq		tmp, vi0, keepspline
	lq.w		tmp, 2(ptr)
	add.y		time, vf0, tmp[w]
	sq.y		time, timeval(vi0)
keepspline:
	isw.x		ptr, splineptr(vi0)
	ilw.w		scene, 0(ptr)
	isw.z		scene, keys(vi0)
	sub			cammat[0], vf0, vf0
	sub			cammat[1], vf0, vf0
	sub			cammat[2], vf0, vf0
	sub			cammat[3], vf0, vf0
	bal			ret, spline
	move.xyz	cammat[3], splinepos
	iaddiu		ptr, ptr, 6
	bal			ret, spline
	sub.xyz		cammat[2], splinepos, cammat[3]
	normalize	cammat[2]
	add.y		cammat[1], vf0, vf0[w]
	opmula.xyz	acc, cammat[1], cammat[2]
	opmsub.xyz	cammat[0], cammat[2], cammat[1]
	normalize	cammat[0]
	opmula.xyz	acc, cammat[0], cammat[2]
	opmsub.xyz	cammat[1], cammat[2], cammat[0]
	
camskipauto:
	
	sq			cammat[3], cameramatrix+3(vi0)
	MatrixTranspose	cammat, cammat
	sq			cammat[0], cameramatrix+0(vi0)
	sq			cammat[1], cameramatrix+1(vi0)
	sq			cammat[2], cameramatrix+2(vi0)

; ****

	iaddiu		packet, vi0, geoinit
	xgkick		packet

	ilw.z		script, keys(vi0)
	iaddiu		script, script, geometriedata
	iaddiu		iterator, vi0, 0
	iaddiu		currentword, vi0, 0
	sub			scriptdata, vf0, vf0
	iaddiu		obuf, vi0, outputbuffers
	iaddiu		obuf2, obuf, 3 * 22
	iaddiu		vnum, vi0, 1
	iaddiu		mstackptr, vi0, 0
	iaddiu		loopsp, vi0, loopstack + 2
	iaddiu		reflectflag, vi0, 0

geoscript_loop:
	; get next script word into scriptdata[x]
	mr32		scriptdata, scriptdata
	ibne		currentword, vi0, geoscript_skipnewword
	lqi			scriptdata, (script++)
geoscript_skipnewword:
	iaddiu		currentword, currentword, 0x4000
geoscript_loopnext:
	
	mtir		opcode, scriptdata[x]
	iaddiu		mask, vi0, 128
	iand		timeanim, opcode, mask
	ibeq		timeanim, vi0, skip_timeanim
	lq.y		time, timeval(vi0)
	mul.x		scriptdata, scriptdata, time[y]
skip_timeanim:
	iaddiu		mask, vi0, 127
	iand		opcode, opcode, mask
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, plane				; opcode 0: plane
	isubiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, settexture		; opcode 1: texture
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, render				; opcode 2: render
	isubiu		opcode, opcode2, 3
	ibeq		opcode2, vi0, tessrepeat		; opcode 3: tessrepeat
	isubiu		opcode2, opcode, 1
	iblez		opcode, move					; opcode 4-6: movex, movey, movez
	isubiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, call				; opcode 7: call
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, return				; opcode 8: return
	isubiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, loop				; opcode 9: loop
	isubiu		opcode2, opcode, 3
	ibeq		opcode, vi0, endloop			; opcode 10: endloop
	isubiu		opcode, opcode2, 4
	iblez		opcode2, rot					; opcode 11-13: rotx, roty, rotz
	isubiu		opcode2, opcode, 3
	iblez		opcode, scale					; opcode 14-17: scale, scalex, scaley, scalez
	isubiu		opcode, opcode2, 1
	iblez		opcode2, scaleuv				; opcode 18-20: scaleu, scalev, scaleuv
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, tube				; opcode 21: tube
	isubiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, cube				; opcode 22: cube
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, reflect			; opcode 23: reflect
	isubiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, disc				; opcode 24: disc
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, alphafunc			; opcode 25: alphafunc
	isubiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, xoffset			; opcode 26: xoffset
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, yoffset			; opcode 27: yoffset
	isubiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, setframe			; opcode 28: setframe
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, blitfrom			; opcode 29: blitfrom
	isubiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, solid				; opcode 30: solid
	isubiu		opcode2, opcode, 1
	ibeq		opcode, vi0, ginit				; opcode 31: init
	
end:
--cont
	b			end

; *** PLANE

plane:
	saveposition
	ftoi0.x		tesselation, scriptdata
	mtir		count, tesselation[x]
	itof0.x		tesselation, tesselation		; get rid of fraction
	div			q, vf0[w], tesselation[x]		; calculate step size
	sub.y		acc, vf0, vf0
	loi			0.5
	sub.xz		acc, vf0, i						; start at [-0.5, 0, -0.5]
	mfir.x		iterator_f, iterator
	itof0.x		iterator_f, iterator_f
	madd.x		acc, iterator_f, q				; go to start pos for this iteration
	madd.xyz	pos, vf0, vf0
	move.yz		pos2, pos
	add.x		pos2, pos, q					; 2nd starting pos
	add.x		uv, pos, i
	add.y		uv, vf0, pos[z]
	add.y		uv, vf0, vf0[w]
	move.y		uv2, uv
	add.x		uv2, uv, q
	sub.xz		normal, vf0, vf0
	add.y		normal, vf0, vf0[w]
	iaddiu		vnum, count, 1
	iadd		vnum, vnum, vnum
	iadd		tmp, vnum, vnum
	iadd		tmp, tmp, vnum					; (tess + 1) * 6
	iaddiu		buffer, vi0, meshbuffer
	iadd		bufend, buffer, tmp
plane_loop:
	sq.xyz		pos, 0(buffer)
	sq.xyz		normal, 1(buffer)
	sq.xy		uv, 2(buffer)
	sq.xyz		pos2, 3(buffer)
	sq.xyz		normal, 4(buffer)
	sq.xy		uv2, 5(buffer)
	add.z		pos, pos, q
	add.z		pos2, pos2, q
	sub.y		uv, uv, q
	sub.y		uv2, uv2, q
	iaddiu		buffer, buffer, 6
	ibne		buffer, bufend, plane_loop
	iaddiu		iterator, iterator, 1
	ibne		iterator, count, initmatrix
	iaddiu		iterator, vi0, 0
	b			initmatrix

; *** TEXTURE
	
settexture:
	ftoi0.x		buffernb, scriptdata
	mtir		buffer, buffernb[x]
	iaddiu		reflectflag, vi0, 0
	bal			ret, set_tex0_to_bufferG
	b			geoscript_loop

; *** RENDER

render:
	lq			giftag, geotag(vi0)
	iaddiu		nloop, vnum, 0x4000
	iaddiu		nloop, nloop, 0x4000			; set EOP bit
	mfir.x		giftag, nloop
	sq			giftag, 0(obuf)
	iaddiu		optr, obuf, 1
	iaddiu		iptr, vi0, meshbuffer
	iadd		tmp, vnum, vnum
	iadd		tmp, tmp, vnum
	iadd		iend, iptr, tmp					; end of input buffer
	lq			clipvec, clipvector(vi0)
	MatrixLoad	omatrix, 0, mstackptr
	MatrixLoad	cammat, cameramatrix, vi0
	sub			matrix[3], omatrix[3], cammat[3]
	MatrixMultiplyVector	matrix[0], cammat, omatrix[0]
	MatrixMultiplyVector	matrix[1], cammat, omatrix[1]
	MatrixMultiplyVector	matrix[2], cammat, omatrix[2]
	MatrixMultiplyVector	matrix[3], cammat, matrix[3]

	ibeq		reflectflag, vi0, skip_reflect
	sub			omatrix[3], omatrix[3], cammat[3]
	add			one, vf0, vf0[w]
.aif 1 eq 0
	lq			inpos, 6(iptr)
	lq			innormal, 7(iptr)
	MatrixMultiplyVector	normal_ip1, omatrix, innormal
	MatrixMultiplyVertexXYZ1	pos_ip1, omatrix, inpos
	mul			dot2, pos_ip1, pos_ip1					; **
	add.x		dot2, dot2, dot2[y]						; **
	add.x		dot2, dot2, dot2[z]						; **
	mul			dot, normal_ip1, pos_ip1				; **
	rsqrt		q, vf0[w], dot2[x]						; **
	move		pos, pos_ip1							; **
	move		normal, normal_ip1

	lq			inpos, 3(iptr)
	lq			innormal, 4(iptr)
	MatrixMultiplyVector	normal_ip1, omatrix, innormal
	MatrixMultiplyVertexXYZ1	pos_ip1, omatrix, inpos

	lq			inpos, 0(iptr)
	lq			innormal, 1(iptr)

.raw
reflect_loop:
	add.x		dot, dot, dot[y]						; *
=	nop
	mul			acc, omatrix[0], innormal[x]			; ***
=	nop
	mul			dot2, pos_ip1, pos_ip1					; **
=	nop
	madd		acc, omatrix[1], innormal[y]			; ***
=	nop
	add.x		dot, dot, dot[z]						; *
=	lq			inpos, 6(iptr)							; ****
	madd		normal_ip2, omatrix[2], innormal[z]		; ***
=	lq			innormals, 7(iptr)						; ****
	add.x		dot2, dot2, dot2[y]						; **
=	loi			0.5
	mul			acc, pos, i								; *
=	nop
	msub		uv, normal, dot[x]						; *
=	nop
	mul			acc, omatrix[0], inpos[x]				; ***
=	move		normal, normal_ip1
	add.x		dot2, dot2, dot2[z]						; **
=	move		pos, pos_ip1							; **
	mul			dot, normal_ip1, pos_ip1				; **
=	move		normal_ip1, normal_ip2					; ***
	mul			uv, uv, q								; *
=	iaddiu		iptr, iptr, 3
	madd		acc, omatrix[1], inpos[y]				; ***
=	rsqrt		q, vf0[w], dot2[x]						; **
	madd		acc, omatrix[2], inpos[z]				; ***
=	ibne		iptr, iend, reflect_loop
	madd		pos_i1p, omatrix[3], vf0[w]				; ***
=	sq.xy		uv, -1(iptr)							; *
.endraw
.aelse
reflect_loop:
;--loopCS 1, 0
	lqi			pos, (iptr++)
	lqi			normal, (iptr++)
	MatrixMultiplyVector	normal, omatrix, normal
	MatrixMultiplyVertexXYZ1	pos, omatrix, pos
	mul			dot, pos, pos
	add.x		dot, dot, dot[y]
	add.x		dot, dot, dot[z]
	rsqrt		q, vf0[w], dot[x]
	mul			dot, normal, pos
	add.x		dot, dot, dot[y]
	add.x		dot, dot, dot[z]
	loi			0.5
	mul			acc, pos, i
	msub		pos, normal, dot[x]
	mul			pos, pos, q
	sqi.xy		pos, (iptr++)

	ibne		iptr, iend, reflect_loop
.aendi
	
	add			omatrix[3], omatrix[3], cammat[3]
	iaddiu		iptr, vi0, meshbuffer
skip_reflect:
	
	loi			512
	add.x		scale, vf0, i
	loi			256
	add.y		scale, vf0, i
	mul.xy		matrix[0], matrix[0], scale
	mul.xy		matrix[1], matrix[1], scale
	mul.xy		matrix[2], matrix[2], scale
	mul.xy		matrix[3], matrix[3], scale
	mul.w		matrix[0], vf0, matrix[0][z]
	mul.w		matrix[1], vf0, matrix[1][z]
	mul.w		matrix[2], vf0, matrix[2][z]
	mul.w		matrix[3], vf0, matrix[3][z]
	add.z		uv, vf0, vf0[w]
	lq			lightvec[0], lights(vi0)
	lq			lightvec[1], lights+1(vi0)
	lq			lightvec[2], lights+2(vi0)
	lq			lights[0], lights+3(vi0)
	lq			lights[1], lights+4(vi0)
	lq			lights[2], lights+5(vi0)
	lq			lights[3], lights+6(vi0)
	; normalize omatrix
	erleng		p, omatrix[0]
	mfp.w		scale, p
	mul.xyz		omatrix[0], omatrix[0], scale[w]
	mul			tmp, omatrix[1], omatrix[1]
	add.x		tmp, tmp, tmp[y]
	add.x		tmp, tmp, tmp[z]
	rsqrt		q, vf0[w], tmp[x]
	mul.xyz		omatrix[1], omatrix[1], q
	mul			tmp, omatrix[2], omatrix[2]
	add.x		tmp, tmp, tmp[y]
	add.x		tmp, tmp, tmp[z]
	rsqrt		q, vf0[w], tmp[x]
	mul.xyz		omatrix[2], omatrix[2], q
renderloop:
--loopCS 1,0
	lqi			pos, (iptr++)
	lqi			normal, (iptr++)
	lqi.xy		uv, (iptr++)
	
	MatrixMultiplyVertexXYZ1	pos, matrix, pos
	
	; culling test
	add.z		acc, vf0, clipvec[w]
	madd.z		clippos, clipvec, pos[z]
	mul.xy		clippos,clipvec, pos
	clipw.xyz	clippos, pos
	fcand		vi1, 0b111111111111111111
	iaddiu		vi1, vi1, 0x7fff
	mfir.w		opos, vi1
	
	div			q, vf0[w], pos[z]
	loi			65536
	add.z		pos, vf0, i
	mul.xyz		pos, pos, q
	loi			2048
	add.xy		pos, pos, i
	ftoi4.xyz	opos, pos

	; lights
	MatrixMultiplyVector	normal, omatrix, normal
	mul			acc, lightvec[0], normal[x]
	madd		acc, lightvec[1], normal[y]
	madd		light, lightvec[2], normal[z]
	max			light, light, vf0[x]
	mul			acc, lights[0], light[x]
	madd		acc, lights[1], light[y]
	madd		acc, lights[2], light[z]
	madd.xzy	rgba, lights[3], light[w]
	ftoi0.xyz	rgba, rgba
	iaddiu		rgba, vi0, 128
	mfir.w		rgba, rgba

	mul.xyz		stq, uv, q

	sqi.xyz		stq, (optr++)
	sqi			rgba, (optr++)
	sqi			opos, (optr++)
	
	ibne		iptr, iend, renderloop
	
	xgkick		obuf
	iadd		tmp, vi0, obuf
	iadd		obuf, vi0, obuf2
	iadd		obuf2,vi0, tmp
	b			geoscript_loop

; *** TESSREPEAT

tessrepeat:
	ibeq		iterator, vi0, geoscript_loop
	lq			scriptdata, loopstack(vi0)
	ilw.x		script, loopstack+1(vi0)
	ilw.y		currentword, loopstack+1(vi0)
	b			geoscript_loopnext

; *** MOVE

move:
	lq			pos, 3(mstackptr)
	iaddiu		opcode2, opcode, 1
	ibeq		opcode, vi0, movez
	ibeq		opcode2, vi0, movey
	add.x		pos, pos, scriptdata[x]
	b			movecont
movey:
	add.y		pos, pos, scriptdata[x]
	b			movecont
movez:
	add.z		pos, pos, scriptdata[x]
movecont:
	sq			pos, 3(mstackptr)
	b			geoscript_loop

; *** CALL

call:
	pushloop
	MatrixLoad	matrix, 0, mstackptr
	iaddiu		mstackptr, mstackptr, 4
	MatrixSave	matrix, 0, mstackptr
	ftoi0.x		tmp, scriptdata
	mtir		script, tmp[x]
	iaddiu		script, script, geometriedata
	iaddiu		currentword, vi0, 0
	sub			scriptdata, vf0, vf0
	b			geoscript_loop

; *** RETURN

return:
	poploop
	isubiu		mstackptr, mstackptr, 4
	b			geoscript_loop

; *** LOOP

loop:
	pushloop
	ftoi0.x		tmp, scriptdata
	mtir		counter, tmp[x]
	isw.z		counter, -1(loopsp)
	b			geoscript_loop

; *** LOOPEND

endloop:
	ilw.z		counter, -1(loopsp)
	isubiu		counter, counter, 1
	isw.z		counter, -1(loopsp)
	ibne		counter, vi0, loopnext
	isubiu		loopsp, loopsp, 2
	b			geoscript_loop
loopnext:
	poploop
	iaddiu		loopsp, loopsp, 2
	b			geoscript_loop

; *** ROT

rot:
	move		angle, scriptdata
	AngleSinCos	angle, sincos
	mr32		rmat[2], vf0
	mr32		rmat[1], rmat[2]
	mr32		rmat[0], rmat[1]
	MatrixLoad	matrix, 0, mstackptr
	iaddiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, rotz
	ibeq		opcode, vi0, roty
	add.y		rmat[1], vf0, sincos[y]
	add.z		rmat[2], vf0, sincos[y]
	add.y		rmat[2], vf0, sincos[x]
	sub.z		rmat[1], vf0, sincos[x]
	b			rotcont
roty:
	add.x		rmat[0], vf0, sincos[y]
	add.z		rmat[2], vf0, sincos[y]
	add.x		rmat[2], vf0, sincos[x]
	sub.z		rmat[0], vf0, sincos[x]
	b			rotcont
rotz:
	add.x		rmat[0], vf0, sincos[y]
	add.y		rmat[1], vf0, sincos[y]
	add.x		rmat[1], vf0, sincos[x]
	sub.y		rmat[0], vf0, sincos[x]
rotcont:
	MatrixMultiplyVector matrix[0], rmat, matrix[0]
	MatrixMultiplyVector matrix[1], rmat, matrix[1]
	MatrixMultiplyVector matrix[2], rmat, matrix[2]
	MatrixMultiplyVector matrix[3], rmat, matrix[3]
	MatrixSave	matrix, 0, mstackptr
	b			geoscript_loop

; *** SCALE

scale:
	add.xyz		scale, vf0, vf0[w]
	MatrixLoad	matrix, 0, mstackptr
	iaddiu		opcode2, opcode, 2
	ibeq		opcode, vi0, scaleall
	ibgtz		opcode2, scalez
	ibeq		opcode2, vi0, scaley
	add.x		scale, vf0, scriptdata[x]
	b			scalecont
scaley:
	add.y		scale, vf0, scriptdata[x]
	b			scalecont
scalez:
	add.z		scale, vf0, scriptdata[x]
	b			scalecont
scaleall:
	add.xyz		scale, vf0, scriptdata[x]
scalecont:
	mul.xyz		matrix[0], matrix[0], scale
	mul.xyz		matrix[1], matrix[1], scale
	mul.xyz		matrix[2], matrix[2], scale
	mul.xyz		matrix[3], matrix[3], scale
	MatrixSave	matrix, 0, mstackptr
	b			geoscript_loop

; *** SCALE UV

scaleuv:
	iaddiu		iptr, vi0, meshbuffer+2
	iadd		iend, vnum, vnum
	iadd		iend, iend, vnum
	iadd		iend, iend, iptr
	add.xy		scale, vf0, vf0[w]
	iaddiu		opcode, opcode2, 1
	ibeq		opcode2, vi0, scaleuvall
	ibeq		opcode, vi0, scalev
	add.x		scale, vf0, scriptdata[x]
	b			scaleuv_cont
scalev:
	add.y		scale, vf0, scriptdata[x]
	b			scaleuv_cont
scaleuvall:
	add.xy		scale, vf0, scriptdata[x]

scaleuv_cont:
	lq.xy		in, 0(iptr)
	mul.xy		out, scale, in
	lq.xy		in, 3(iptr)
.raw
scaleuv_loop:
	mul.xy		acc, vf0, vf0
=	iaddiu		iptr, iptr, 3
	madd.xy		acc, scale, in
=	lq.xy		in, 3(iptr)
	nop
=	ibne		iptr, iend, scaleuv_loop
	madd.xy		out, vf0, vf0
=	sq.xy		out, -3(iptr)
.endraw
	b			geoscript_loop

; *** TUBE

tube:
	saveposition
	ftoi0.x		tesselation, scriptdata
	mtir		count, tesselation[x]
	itof0.x		tesselation, tesselation		; get rid of fraction
	div			q, vf0[w], tesselation[x]		; calculate step size
	sub.z		acc, vf0, vf0
	loi			0.5
	add.y		acc, vf0, i
	sub.x		acc, vf0, i						; start at [-0.5, 0.5, 0]
	mfir.x		iterator_f, iterator
	itof0.x		iterator_f, iterator_f
	madd.x		acc, iterator_f, q				; go to start pos for this iteration
	madd.xyz	pos, vf0, vf0
	move.yz		pos2, pos
	add.x		pos2, pos, q					; 2nd starting pos
	add.x		uv, pos, i
	add.y		uv, vf0 ,uv[x]
	sub.x		uv, vf0, vf0
	move.x		uv2, uv
	add.y		uv2, uv, q
	sub.xz		normal, vf0, vf0
	add.y		normal, vf0, vf0[w]
	iaddiu		vnum, count, 1
	iadd		vnum, vnum, vnum
	iadd		tmp, vnum, vnum
	iadd		tmp, tmp, vnum					; (tess + 1) * 6
	iaddiu		buffer, vi0, meshbuffer
	iadd		bufend, buffer, tmp
	loi			0.766044						; cos
	add.y		sincos, vf0, i
	add.z		cossin, vf0, i
	loi			0.642788					; sin
	add.z		sincos, vf0, i
	sub.y		cossin, vf0, i
tube_loop:
	sq.xyz		pos, 0(buffer)
	sq.xyz		normal, 1(buffer)
	sq.xy		uv, 2(buffer)
	sq.xyz		pos2, 3(buffer)
	sq.xyz		normal, 4(buffer)
	sq.xy		uv2, 5(buffer)
	
	mul.yz		acc, sincos, pos[y]
	madd.yz		pos, cossin, pos[z]
	madd.yz		pos2, cossin, pos2[z]
	mul.yz		acc, sincos, normal[y]
	madd.yz		normal, cossin, normal[z]
	add.x		uv, uv, q
	add.x		uv2, uv2, q
	iaddiu		buffer, buffer, 6
	ibne		buffer, bufend, tube_loop
	iaddiu		iterator, iterator, 1
	ibne		iterator, count, initmatrix
	iaddiu		iterator, vi0, 0
	b			initmatrix

; *** CUBE
	
cube:
	saveposition
	mfir.xyz	iterator_f, iterator
	itof0.xyz	iterator_f, iterator_f
	sub.xyz		iterator_f, iterator_f, vf0[w]	; scale factor for this half of the cube
	mul.xyz		scale, iterator_f, scriptdata[x]
	iaddiu		buffer, vi0, meshbuffer
	iaddiu		vnum, vi0, 12
	iaddiu		positions, vi0, cubeverts
	iaddiu		normals, vi0, cubenormals
	iaddiu		bufferend, buffer, 36
	loi			0.5
	mul.xy		tmp, iterator_f, i
	sub.xy		uv0, vf0, tmp
	sub.y		uv1, vf0, tmp
	sub.x		uv2, vf0, tmp
	add.x		uv1, vf0, tmp
	add.y		uv2, vf0, tmp
	add.xy		uv3, vf0, tmp
	
	sub.xy		uv0, vf0, vf0
	sub.y		uv1, vf0, vf0
	sub.x		uv2, vf0, vf0
	add.x		uv1, vf0, vf0[w]
	add.y		uv2, vf0, vf0[w]
	add.xy		uv3, vf0, vf0[w]
cube_loop:
	lqi.xyz		normal, (normals++)
	mul.xyz		normal, normal, iterator_f
	lqi.xyz		pos, (positions++)
	mul.xyz		pos, pos, scale
	sq.xyz		pos, 0(buffer)
	mul.xyz		pos, pos, scale
	sq.xyz		normal, 1(buffer)
	sq.xy		uv0, 2(buffer)
	lqi.xyz		pos, (positions++)
	mul.xyz		pos, pos, scale
	sq.xyz		pos, 3(buffer)
	sq.xyz		normal, 4(buffer)
	sq.xy		uv1, 5(buffer)
	lqi.xyz		pos, (positions++)
	mul.xyz		pos, pos, scale
	sq.xyz		pos, 6(buffer)
	sq.xyz		normal, 7(buffer)
	sq.xy		uv2, 8(buffer)
	lqi.xyz		pos, (positions++)
	mul.xyz		pos, pos, scale
	sq.xyz		pos, 9(buffer)
	sq.xyz		normal, 10(buffer)
	sq.xy		uv3, 11(buffer)
	iaddiu		buffer, buffer, 12
	ibne		buffer, bufferend, cube_loop
	iadd		tmp, vi0, iterator
	iaddiu		iterator, iterator, 2
	ibeq		tmp, vi0, initmatrix
	iaddiu		iterator, vi0, 0
	b			initmatrix

; *** reflect

reflect:
	iaddiu		reflectflag, vi0, 1
	b			geoscript_loop
; *** disc

disc:
	saveposition
	iaddiu		buffer, vi0, meshbuffer
	ftoi0.x		tesselate, scriptdata
	mtir		vnum, tesselate[x]
	iaddiu		vnum, vi0, 9
	iaddiu		vnum, vnum, 1
	iadd		vnum, vnum, vnum
	iadd		bufend, vnum, vnum
	iadd		bufend, bufend, vnum
	iadd		bufend, bufend, buffer
	sub			normal, vf0, vf0
	add.y		normal, vf0, vf0[w]
	sub			pos0, vf0, vf0
	sub			pos1, vf0, vf0
	loi			0.5
	add.y		pos0, vf0, i
	add.xy		uv1, vf0, i
	loi			0.766044						; cos
	add.x		sincos, vf0, i
	add.y		cossin, vf0, i
	loi			0.642788					; sin
	add.y		sincos, vf0, i
	sub.x		cossin, vf0, i
disc_loop:
	add.xy		uv0, pos0, i
	sqi			pos0, (buffer++)
	sqi			normal, (buffer++)
	sqi.xy		uv0, (buffer++)
	sqi			pos1, (buffer++)
	sqi			normal, (buffer++)
	sqi.xy		uv1, (buffer++)
	mul.xy		acc, sincos, pos0[x]
	madd.xy		pos0, cossin, pos0[y]
	ibne		buffer, bufend, disc_loop
	iaddiu		iterator, vi0, 0
	b			initmatrix

; *** alphafunc

alphafunc:
	iaddiu		packet, vi0, free_mem
	lq			giftag, ADtag(vi0)
	sq			giftag, 0(packet)
	ftoi0.x		scriptdata, scriptdata
	mtir		funcnr, scriptdata[x]
	lq			func, alphafunctions(funcnr)
	sq			func, 1(packet)
	xgkick		packet
	xgkick		vi0
	lq			giftag, colortag(vi0)
	sq			giftag, 0(packet)
	mtir		alpha, func[w]
	isw.xyzw	alpha, 1(packet)
	xgkick		packet
	xgkick		vi0
	b			geoscript_loop

; *** xoffset

xoffset:
	sq.x		scriptdata, blitoffset(vi0)
	b			geoscript_loop

; *** y offset

yoffset:
	add.y		tmp, vf0, scriptdata[x]
	sq.y		tmp, blitoffset(vi0)
	b			geoscript_loop

; *** setframe

setframe:
	bal			ret, getrealbuffernr
	iadd		buffer, buffer, buffer
	iaddiu		packet, buffer, framepackets
	xgkick		packet
	b			geoscript_loop

; *** blitfrom

blitfrom:
	bal			ret, getrealbuffernr
	iadd		buffer, buffer, buffer
	iaddiu		packet, buffer, tex0packets
	xgkick		packet
	iaddiu		packet, vi0, texgeninitpacket
	xgkick		packet
	iaddiu		packet, vi0, fxinit
	xgkick		packet
	iaddiu		packet, vi0, free_mem
	lq			giftag, spritestqgiftag(vi0)
	sq			giftag, 0(packet)
	lq			uv, blitoffset(vi0)
	add.z		uv, vf0, vf0[w]
	loi			1.0/1024.0
	add.x		uvscale, vf0, i
	loi			1.0/256.0
	add.y		uvscale, vf0, i
	add.z		uvscale, vf0, vf0[w]
	loi			32
	add.x		step1, vf0, i
	loi			224
	add.y		step1, vf0, i
	sub			pos, vf0, vf0
	iaddiu		cnt, vi0, 20
blitloop:
	mul.xyz		uvi, uv, uvscale
	sq.xyz		uvi, 1(packet)
	ftoi4		posi, pos
	sq			posi, 2(packet)
	add.xy		pos, pos, step1
	add.xy		uv, uv, step1
	mul.xyz		uvi, uv, uvscale
	sq.xyz		uvi, 3(packet)
	ftoi4		posi, pos
	sq			posi, 4(packet)
	sub.y		pos, pos, i
	sub.y		uv, uv, i
	xgkick		packet
	xgkick		vi0
	isubiu		cnt, cnt, 1
	ibne		cnt, vi0, blitloop
	b			geoscript_loop

; *** SOLID

solid:
	iaddiu		packet, vi0, texgeninitpacket
	xgkick		packet
	iaddiu		packet, vi0, free_mem
	lq			giftag, colortag(vi0)
	sq			giftag, 0(packet)
	ftoi0.x		scriptdata, scriptdata
	mtir		alpha, scriptdata[x]
	isw.xyzw	alpha, 1(packet)
	xgkick		packet
	xgkick		vi0
	lq			giftag, solidsprite(vi0)
	sq			giftag, 0(packet)
	sub			pos, vf0, vf0
	loi			32
	add.x		step1, vf0, i
	loi			224
	add.y		step1, vf0, i
	iaddiu		cnt, vi0, 20
solidloop:
	ftoi4		posi, pos
	sq			posi, 1(packet)
	add.xy		pos, pos, step1
	ftoi4		posi, pos
	sq			posi, 2(packet)
	sub.y		pos, pos, i
	xgkick		packet
	xgkick		vi0
	isubiu		cnt, cnt, 1
	ibne		cnt, vi0, solidloop
	b			geoscript_loop

; *** INIT

ginit:
	iaddiu		packet, vi0, geoinit
	xgkick		packet
	b			geoscript_loop
	
initmatrix:
	iaddiu		mstackptr, vi0, matrixstack
	sub			mat, vf0, vf0
	sq			mat, 3(mstackptr)
	add.z		mat, vf0, vf0[w]
	sq			mat, 2(mstackptr)
	mr32		mat, mat
	sq			mat, 1(mstackptr)
	mr32		mat, mat
	sq			mat, 0(mstackptr)
	b			geoscript_loop

getrealbuffernr:
	ftoi0.x		scriptdata, scriptdata
	mtir		buffer, scriptdata[x]
	iaddiu		tmp, vi0, 2
	ibeq		buffer, tmp, grbn_end
	ilw.x		tmp, 0x3ff(vi0)
	iaddiu		tmp2, vi0, 1
	ibne		tmp, vi0, grbn_end
	isub		buffer, tmp2, buffer
grbn_end:
	jr			ret
	
set_tex0_to_bufferG:
	; in: buffer (i) - buffer to set
	iaddiu		tmp, vi0, free_mem
	lq			tmp2, buftex0packet(vi0)
	sq			tmp2, 0(tmp)
	lq			tmp2, buftex0packet+1(vi0)
	sq			tmp2, 1(tmp)
	mfir.x		tmp3, buffer
	itof0.x		tmp3, tmp3
	loi			0Xd4800000
	mul.x		tmp3, tmp3, i
	add.x		tmp2, tmp2, tmp3
	sq.x		tmp2, 1(tmp)
	xgkick		tmp
	xgkick		vi0
	jr			ret

spline:
	; in: ptr
	;     t[x]
	; out: splinepos
	loi			0x800000
	add.xyz		one, vf0, vf0[w]
	ftoi0.x		t1, t
	mtir		tptr, t1[x]
	iadd		tptr, tptr, ptr
	add.xyz		acc, vf0, t[x]
	madd.xyz	acc, one, i
	msub.xyz	t2, one, i
	add.xyz		t1, vf0, t[x]
	sub.xyz		t1, t1, t2
	lq.xyz		p0, 1(tptr)
	lq.xyz		p1, 0(tptr)
	lq.xyz		p2, 3(tptr)
	lq.xyz		p3, 2(tptr)
	sub.xyz		v1, p3, p1
	sub.xyz		v2, p0, p2
	loi			1.0 / 6.0
	add.xyz		acc, p0, vf0
	madd.xyz	p1, v1, i
	add.xyz		acc, p3, vf0
	madd.xyz	p2, v2, i
	mul.xyz		t2, t1, t1
	mul.xyz		t3, t1, t2
	sub.xyz		nt1, one, t1
	mul.xyz		nt2, nt1, nt1
	mul.xyz		nt3, nt1, nt2
	mul.xyz		acc, nt3, p0
	madd.xyz	acc, t3, p3
	loi			3.0
	mul.xyz		tmp, nt2, t1
	mul.xyz		p1, p1, i
	madd.xyz	acc, tmp, p1
	mul.xyz		tmp, nt1, t2
	mul.xyz		p2, p2, i
	madd.xyz	splinepos, tmp, p2
	jr			ret

--exitm
--endexit

.END
