;
; tree.vcl - 7-apr-2003
;
; By Jon Rocatis - jr@ioi.dk
;
; Renders a procedurally generated tree
; TAB size 4
;
; Thanks to Morten Mikkelsen for moral support and bringing me 
; up to speed on various issues :)
;

; Last qword layout: [drawbuffer, drawbuffer, RandomFloat, buttons]

;-------------------------------------------------------------------------------------------

POINTS		.assign	6

PRIMSIZESLOT	.assign (((POINTS+1)*2*2)+1)		; incl gif
PRIMSIZEBANK	.assign (PRIMSIZESLOT*2)
PRIMSIZETOTAL	.assign	(PRIMSIZEBANK*2)

STACKFRAMESIZE	.assign	19				;quads

LASTQW		.assign		1023		; end of memory

JOY_START			.assign		8			; 0x0008
JOY_UP				.assign		16			; 0x0010
JOY_RIGHT			.assign		32			; 0x0020
JOY_DOWN			.assign		64			; 0x0040
JOY_LEFT			.assign		128			; 0x0080
JOY_TRIANGLE		.assign		4096		; 0x1000
JOY_CIRCLE			.assign		8192		; 0x2000
JOY_X				.assign		16384		; 0x4000
JOY_SQUARE			.assign		32767		; 0x7FFF
JOY_R2				.assign		512			; 0x0200
JOY_R1				.assign		2048		; 0x0800

	.include	"vcl_sml.i"
	.include	"mymacros.i"
	.include	"data.i"

;-------------------------------------------------------------------------------------------

.init_vf_all
.init_vi_all
.syntax new

.vu

--enter
--endenter

START:
	ilw.w		itmp,SETTINGS+7(vi00)
	ibne		itmp,vi00,notinit

	iaddiu		itmp,vi00,TEXUPLOAD
	xgkick		itmp
	iaddiu		itmp,vi00,1
	isw.w		itmp,SETTINGS+7(vi00)
	b			alldone

notinit:
	iaddiu		background,vi00,REGSETUP
	xgkick		background

	iaddiu		background,vi00,BACKGROUND
	xgkick		background

	iaddiu		stackptr,vi00,1023-STACKFRAMESIZE		; top of mem
	iaddiu		level,vi00,0
	iaddiu		primbanknum,vi00,0						; current

	; Punch in giftags in both prim banks
	iaddiu		gifptr,vi00,PRIMBUFFER
	lq			tmp,SETTINGS+6(vi00)					; Get gif tag end
	sq			tmp,0(gifptr)
	sq			tmp,PRIMSIZESLOT(gifptr)
	sq			tmp,PRIMSIZEBANK+0(gifptr)
	sq			tmp,PRIMSIZEBANK+PRIMSIZESLOT(gifptr)

	;--------------------------------------------------------------------------------------------------------------------------
	; Joypad handling

	ilw.w			itmp,1023(vi00)				; Get buttons

	;-------------
	; Handle twist
	iaddiu			imask,vi00,JOY_X
	iand			imask,itmp,imask
	ibne			imask,vi00,nobut_x

	lq.w			tmp,SETTINGS+1(vi00)
	loi				0x3C23D70A
	addi.x			tmp2,vf00,i

	iaddiu			imask,vi00,JOY_UP
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_twist_1
	add.w			tmp,tmp,tmp2[x]
joy_twist_1:
	iaddiu			imask,vi00,JOY_DOWN
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_twist_2
	sub.w			tmp,tmp,tmp2[x]
joy_twist_2:
	sq.w			tmp,SETTINGS+1(vi00)
nobut_x:

	;-------------
	; Handle threshhold

	iaddiu			imask,vi00,JOY_R2
	iand			imask,itmp,imask
	ibne			imask,vi00,nobut_r2

	lq.z			tmp,SETTINGS+0(vi00)
	loi				0x3A83126F
	addi.x			tmp2,vf00,i

	iaddiu			imask,vi00,JOY_UP
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_threshhold_1
	sub.z			tmp,tmp,tmp2[x]

joy_threshhold_1:
	iaddiu			imask,vi00,JOY_DOWN
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_threshhold_2
	add.z			tmp,tmp,tmp2[x]

joy_threshhold_2:
	loi				0x3F000000
	addi.y			tmp,vf00,i
	max.z			tmp,tmp,vf00[x]					; clamp to zero
	mini.z			tmp,tmp,tmp[y]					; clamp to 0.5
	sq.z			tmp,SETTINGS+0(vi00)

nobut_r2:

	;-------------
	; Handle spread

	iaddiu			imask,vi00,0x4000
	iaddiu			imask,imask,0x4000				; square
	iand			imask,itmp,imask
	ibne			imask,vi00,nobut_sqr

	lq.z			tmp,SETTINGS+1(vi00)
	loi				0x3C23D70A
	addi.x			tmp2,vf00,i

	iaddiu			imask,vi00,JOY_UP
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_spread_1
	sub.z			tmp,tmp,tmp2[x]
joy_spread_1:
	iaddiu			imask,vi00,JOY_DOWN
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_spread_2
	add.z			tmp,tmp,tmp2[x]
joy_spread_2:

	max.z			tmp,tmp,vf00[x]					; clamp to zero
	mini.z			tmp,tmp,vf00[w]					; clamp to one
	sq.z			tmp,SETTINGS+1(vi00)

nobut_sqr:

	;-------------
	; Handle branch angle

	iaddiu			imask,vi00,JOY_R1
	iand			imask,itmp,imask
	ibne			imask,vi00,nobut_r1

	lq.y			tmp,SETTINGS+1(vi00)
	loi				0x3C23D70A
	addi.x			tmp2,vf00,i

	iaddiu			imask,vi00,JOY_UP
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_angle_1
	add.y			tmp,tmp,tmp2[x]
joy_angle_1:
	iaddiu			imask,vi00,JOY_DOWN
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_angle_2
	sub.y			tmp,tmp,tmp2[x]
joy_angle_2:

	max.y			tmp,tmp,vf00[x]					; clamp to zero
	mini.y			tmp,tmp,vf00[w]					; clamp to one

	sq.y			tmp,SETTINGS+1(vi00)

nobut_r1:

	;-------------
	; Handle leaf size

	iaddiu			imask,vi00,JOY_TRIANGLE
	iand			imask,itmp,imask
	ibne			imask,vi00,nobut_tri1

	lq.y			tmp,SETTINGS+0(vi00)
	loi				0x3C23D70A
	addi.x			tmp2,vf00,i

	iaddiu			imask,vi00,JOY_UP
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_leaf_1
	add.y			tmp,tmp,tmp2[x]
joy_leaf_1:
	iaddiu			imask,vi00,JOY_DOWN
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_leaf_2
	sub.y			tmp,tmp,tmp2[x]
joy_leaf_2:

	loi				0x40A00000
	addi.x			bound,vf00,i
	max.y			tmp,tmp,vf00[x]					; clamp to zero
	mini.y			tmp,tmp,bound[x]				; clamp upper
	sq.y			tmp,SETTINGS+0(vi00)

nobut_tri1:

	;-------------
	; Handle branch thickness

	iaddiu			imask,vi00,JOY_CIRCLE
	iand			imask,itmp,imask
	ibne			imask,vi00,nobut_c1

	lq.w			tmp,SETTINGS+0(vi00)
	loi				0x3BA3D70A					; 0.005 3C23D70A
	addi.x			tmp2,vf00,i

	iaddiu			imask,vi00,JOY_UP
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_thickness_1
	add.w			tmp,tmp,tmp2[x]
joy_thickness_1:
	iaddiu			imask,vi00,JOY_DOWN
	iand			imask,itmp,imask
	ibne			imask,vi00,joy_thickness_2
	sub.w			tmp,tmp,tmp2[x]
joy_thickness_2:

	loi				0x3D23D70A				; 0.04 3D4CCCCD
	addi.x			bounds,vf00,i
	loi				0x3E4CCCCD
	addi.y			bounds,vf00,i

	max.w			tmp,tmp,bounds[x]
	mini.w			tmp,tmp,bounds[y]
	sq.w			tmp,SETTINGS+0(vi00)

nobut_c1:

	;-------------
	; Handle default

	iaddiu			imask,vi00,JOY_START
	iand			imask,itmp,imask
	ibne			imask,vi00,nobut_start

	lq				tmp,SETTINGS+3(vi00)
	sq				tmp,SETTINGS+0(vi00)
	lq				tmp,SETTINGS+4(vi00)
	sq				tmp,SETTINGS+1(vi00)

	lq.z			tmp,LASTQW(vi00)
	mr32.y			tmp,tmp
	sq.y			tmp,SETTINGS+7(vi00)

nobut_start:

	;---------------------------------------------------------------------------------------------------------------------------------------

	;-----------------
	; init random

	lq.y		tmp,SETTINGS+7(vi00)					; get random seed
	rinit		r,tmp[y]

	;-----------------
	; precalc radius

	lq.y		leftradius,SETTINGS+1(vi00)				; branch_angle
	lq.y		rightradius,SETTINGS+1(vi00)
	sub.w		leftradius,vf00,leftradius[y]			; 1 - branch_angle
	sqrt		q,leftradius[w]
	addq.z		leftradius,vf00,q
	sq.z		leftradius,SETTINGS+2(vi00)				; sqrtf(1.0 - branch_angle)
	sqrt		q,rightradius[y]
	mulq.w		rightradius,vf00,q
	sq.w		rightradius,SETTINGS+2(vi00)			; sqrtf(branch_angle)

	;-----------------
	; precalc angles

	lq.z		b_anglebias,SETTINGS+1(vi00)
	loi			0x3FC90FDB								; pi/2
	addi.x		tmp,vf00,i
	mul.x		tmp,tmp,b_anglebias[z]					; (3.141592653589f / 2) * tree.branch_angle_bias

	lq.y		b_angle,SETTINGS+1(vi00)
	mr32.x		b_angle,b_angle							; x <- y
	add.y		b_angle,vf00,vf00[w]					; y = 1
	sub.y		b_angle,b_angle,b_angle[x]				; 1 - branch_angle
	mul.xy		b_angle,b_angle,tmp[x]
	sub.y		b_angle,vf00,b_angle[y]					; negate y
	sq.xy		b_angle,SETTINGS+2(vi00)				; precalc_leftangle, precalc_rightangle

	;-- rotation mtx

	ilw.w		itmp,1023(vi00)				; Get buttons
	iaddiu		imask,vi00,JOY_LEFT
	iand		imask,itmp,imask
	ibeq		imask,vi00,pauserot

	MatrixLoad	trans,TRANSMTX,vi00
	lq.x		objrot,SETTINGS+7(vi00)
	move.z		objrot,vf00
	AngleSinCos	objrot,rotsincos
	loi			0x3C23D70A
	addi.y		objrot,vf00,i
	add.x		objrot,objrot,objrot[y]
	sq.x		objrot,SETTINGS+7(vi00)

	add.x		trans[0],vf00,rotsincos[y]
	add.z		trans[0],vf00,rotsincos[x]
	sub.x		trans[2],vf00,rotsincos[x]
	add.z		trans[2],vf00,rotsincos[y]

	MatrixSave	trans,TRANSMTX,vi00
pauserot:

	MatrixIdentity mtx
	add.x		decay,vf00,vf00[w]						; decay = 1

	//------- End of init -----------------------------

	//---------------------------------------------------------------------------
	//-- start of recursive stuff -----------------------------------------------
	//---------------------------------------------------------------------------

dorecur:
	rnext.x		radius,r
	rnext.y		radius,r
	sub.xy		radius,radius,vf00[w]
	loi			0x3E99999A						; 0.3 (static)
	addi.x		tmp,vf00,i
	loi			0x3DCCCCCD						; 0.1 (static)
	addi.y		tmp,vf00,i
	mul.xy		radius,radius,tmp[x]
	sub.xy		radius,radius,tmp[y]
	lq.x		bnoise,SETTINGS+1(vi00)			; get branch noise
	mul.xy		radius,radius,bnoise[x]

	lq.zw		precalcradius,SETTINGS+2(vi00)
	add.x		radius,radius,precalcradius[z]
	add.y		radius,radius,precalcradius[w]

	; clamp left and right radius to [0;1] range
	max.xy		radius,radius,vf00[x]					; clamp to zero
	mini.xy		radius,radius,vf00[w]					; clamp to one

	; here radius[xy] is radius left,right
	
	add.x		upper_radius,vf00,radius[y]			; move y to x (right_radius to x)
	max.x		upper_radius,upper_radius,radius	; max(left_radius, right_radius)

	lq.w		radius,SETTINGS+0(vi00)
	mul.x		upper_radius,upper_radius,radius[w]		; radius * max(left_radius, right_radius);

	iaddiu		cnt,vi0,POINTS+1
	iaddiu		vtxptr,vi00,VTX

	iaddiu		storeptr,vi00,PRIMBUFFER+1					; skip gif tag
	ibeq		primbanknum,vi00,primbank0
	iaddiu		storeptr,vi00,PRIMBUFFER+PRIMSIZEBANK+1		; skip gif tag
primbank0:
	iadd		stitchsource,vi00,storeptr

	lq.w		radius,SETTINGS+0(vi00)
	mr32.z		radius,radius
	sub.w		radius,vf00,radius[w]				; # # radius, 1-radius

	MatrixLoad		trans,TRANSMTX,vi00
	MatrixLoad		perspmtx,PERSPMTX,vi00

	MatrixMultiply	mtxtmp,mtx,trans				; trans x mtx
	MatrixMultiply	bigmtx,mtxtmp,perspmtx			; Actually (P x T)!!

	MatrixLoad4x3		lightdirs,(SETTINGS+8),vi00
	MatrixMultiply4x3   mtxtmp,lightdirs,mtxtmp
	MatrixLoad4x3		lightcols,(SETTINGS+11),vi00

	div			q,vf00[w],decay[x]
	addq.x		oneoverdecay,vf00,q					; for getting normals back to unit length

	; ------- Main cylinder loop -----------------------------------------------------------------------------

	iaddiu		vtxsave,stackptr,5					; we need to store it stack also
	move.w		tmpcol,vf00

vtxloop:
	lq.xz		vtxl,0(vtxptr)
	lqi.xz		vtxu,(vtxptr++)

	move.y		normal,vf00
	mul.xz		normal,vtxl,oneoverdecay[x]			; vector will be unit length after mtx mul
	MatrixMultiplyVector	normal,mtxtmp,normal
	max.xyz		normal,normal,vf00[x]				; clamp to zero
	mul.xyz		tmpcol,lightcols[0],normal[x]		; intensity for light0 [0;1]

	loi			0x437F0000
	addi.x		lort,vf00,i
	mr32.w		lort,lort
	move.w		tmpcol,lort
	ftoi0.xyzw	tmpcol,tmpcol

	mul.xz		vtxl,vtxl,radius[z]					; * tree.branch_radius
	mul.xz		vtxu,vtxu,upper_radius[x]
	add.y		vtxl,vf00,radius[z]
	add.y		vtxu,vf00,radius[w]

	MatrixMultiplyVertexW1	vtxl,bigmtx,vtxl
	MatrixMultiplyVertexW1	vtxu,bigmtx,vtxu

	div			q,vf00[w],vtxl[w]
	mul.xyz		vtxl,vtxl,q

	div			q,vf00[w],vtxu[w]
	mul.xyz		vtxu,vtxu,q

	mr32.w		vtxl,vf00						; clear w
	mr32.w		vtxu,vf00						; clear w

	ftoi4.xyz	vtxl,vtxl
	ftoi4.xyz	vtxu,vtxu

	sq			tmpcol,0(storeptr)				; Store col
	sq.xyzw		vtxl,1(storeptr)				; Store coord

	sq			tmpcol,2(storeptr)				; Store col
	sq.xyzw		vtxu,3(storeptr)				; Store coord

	sq			tmpcol,0(vtxsave)				; Store col
	sq.xyzw		vtxu,1(vtxsave)					; Store coord

	iaddiu		vtxsave,vtxsave,2
	iaddiu		storeptr,storeptr,4

	isubiu		cnt,cnt,1
	ibne		cnt,vi00,vtxloop

	; -- end of main cylinder vtx loop -------------------------------------------------------------------------

	ibeq		level,vi00,nostitch				; Nothing to stich for level 0

	; -- stich ---------------------------------------------------------------------------
	; The stich polys just follow the main ones..
	; L = Copy from stack
	; U = Copy from cur. L (main cylinder)

	; change first gif header (will never stop again)
	lq			tmp,SETTINGS+5(vi00)			; Get gif tag
	sq			tmp,-PRIMSIZESLOT(storeptr)
	iaddiu		storeptr,storeptr,1				; Skip gif tag
	iaddiu		cnt,vi00,POINTS+1

	; calc source ptrs
	iadd		source_u,vi00,stitchsource
	iaddiu		source_l,stackptr,STACKFRAMESIZE+5

stitchloop:
	lq			tmpu0,2(source_u)
	lq			tmpu1,3(source_u)
	lq			tmpl0,0(source_l)
	lq			tmpl1,1(source_l)
	iaddiu		source_u,source_u,4
	iaddiu		source_l,source_l,2
	sq			tmpu0,0(storeptr)			; lower
	sq			tmpu1,1(storeptr)
	sq			tmpl0,2(storeptr)			; upper
	sq			tmpl1,3(storeptr)
	iaddiu		storeptr,storeptr,4
	isubiu		cnt,cnt,1
	ibne		cnt,vi00,stitchloop
nostitch:

	; --- stitching done ---------------------------------------------------------------------------

	iaddiu		primdata,vi00,PRIMBUFFER
	ibeq		primbanknum,vi00,primbankp0
	iaddiu		primdata,vi00,PRIMBUFFER+PRIMSIZEBANK
primbankp0:

	xgkick		primdata						; Draw branch (and stitch polys if any)
	iaddiu		flip,vi00,1
	isub		primbanknum,flip,primbanknum	; toggle storage bank

	; apply translation to current mtx (Translate 1 unit (will be scaled) along cur rot) (actually: translate(0.0f, 1.0f, 0.0f) * mat
	add.xyz		mtx[3],mtx[1],mtx[3]
	MatrixSave	mtx,0,stackptr					; push it on stack (right branch needs it)

	;------ Left branch -------------------------------------------------------------------------------

	lq.w		ftwist,SETTINGS+1(vi00)			; get static twist
	lq.x		left_angle,SETTINGS+2(vi00)		; left_angle
	iaddiu		itmp,level,1
	mfir.w		tmp,itmp
	itof0.w		tmp,tmp
	mul.w		ftwist,ftwist,tmp				; Calc twist

	sub.z		ftwist,vf00,left_angle[x]		; -angle
	add.x		ftwist,vf00,ftwist[w]			; x = w

	AngleSinCos	ftwist,sincos					; input(xz), output	(Calc sincos for twist and left_angle)
	move.x		axis,sincos
	add.z		axis,vf00,sincos[y]				; axis(s,0,c)

	LoadRotationAxisAngleY0 mat,axis,sincos		; axis(s,0,c), -left_angle

	mul.xyz		mat[0],mat[0],radius[x]			; apply scale to rot mtx
	mul.xyz		mat[1],mat[1],radius[x]			; apply scale to rot mtx
	mul.xyz		mat[2],mat[2],radius[x]			; apply scale to rot mtx

	;  left_mat = rot_mat * scale_mat;
	;  left_mat = translate_mat * left_mat;
	; -- here rot_mat and scale_mat are done and multiplied

	MatrixMultiply	left_mat,mat,mtx			; Actually (mtx x mat)!!
	MatrixCopy		mtx,left_mat				; this "transfers" it to next recursion

	sq.x		decay,4(stackptr)				; push decay (done here because right branch always read from stack)
	sq.y		radius,4(stackptr)				; push right_radius

	ilw.x		idepth,SETTINGS+0(vi00)			; Get tree depth
	isubiu		idepth,idepth,1
	isub		itmp,level,idepth
	ibgez		itmp,doleafl					; ((level + 1) >= tree.depth)

	lq.z		threshold,SETTINGS+0(vi00)
	mul.x		tmp,radius,decay[x]				; left_decay
	iaddiu		icmp,vi00,128					; SignX bit
	sub.x		tmp,tmp,threshold[z]			; left_decay - threshold
	fmand		icmp,icmp
	ibne		vi00,icmp,doleafl

dobranchl:
	mul.x		decay,radius,decay[x]					; left_decay
	isw.z		vi00,4(stackptr)
	isubiu		stackptr,stackptr,STACKFRAMESIZE
	iaddiu		level,level,1
	b			dorecur

	;-------------------------------------------------------------------------------------------

doleafl:
	iaddiu		ileaftype,vi00,0
	b			createleaf

leftdone:

	;------ Right branch -------------------------------------------------------------------------------

	; Grab saved data from stack
	lq.x		decay,4(stackptr)				; get decay
	lq.y		radius,4(stackptr)				; get right_radius

	MatrixLoad	mtx,0,stackptr					; get matrix with translation applied

	lq.w		ftwist,SETTINGS+1(vi00)			; get static twist
	lq.y		right_angle,SETTINGS+2(vi00)	; right_angle
	iaddiu		itmp,level,1
	mfir.w		tmp,itmp
	itof0.w		tmp,tmp
	mul.w		ftwist,ftwist,tmp				; Calc twist

	sub.z		ftwist,vf00,right_angle[y]		; -angle
	add.x		ftwist,vf00,ftwist[w]			; x <- w

	AngleSinCos	ftwist,sincos					; input(xz), output	(Calc sincos for twist and right_angle)
	move.x		axis,sincos
	add.z		axis,vf00,sincos[y]				; axis(s,0,c)

	LoadRotationAxisAngleY0 mat,axis,sincos		; axis(s,0,c), -left_angle

	mul.xyz		mat[0],mat[0],radius[y]			; apply scale to rot mtx
	mul.xyz		mat[1],mat[1],radius[y]			; apply scale to rot mtx
	mul.xyz		mat[2],mat[2],radius[y]			; apply scale to rot mtx

	MatrixMultiply	right_mat,mat,mtx			; Actually (mtx x mat)!!
	MatrixCopy		mtx,right_mat				; this "transfers" it to next recursion

	ilw.x		idepth,SETTINGS+0(vi00)			; Get tree depth
	isubiu		idepth,idepth,1
	isub		itmp,level,idepth
	ibgez		itmp,doleafr					; ((level + 1) >= tree.depth)

	lq.z		threshold,SETTINGS+0(vi00)
	mul.y		decay,radius,decay[x]			; right_decay
	iaddiu		icmp,vi00,64					; SignY bit
	sub.y		tmp,decay,threshold[z]			; right_decay - threshold
	fmand		icmp,icmp
	ibne		vi00,icmp,doleafr
	
dobranchr:
	mul.y		decay,radius,decay[x]
	mr32.x		decay,decay						; copy it to real decay
	iaddiu		itmp,vi00,1
	isw.z		itmp,4(stackptr)
	isubiu		stackptr,stackptr,STACKFRAMESIZE
	iaddiu		level,level,1
	b			dorecur							; recurse

	;-------------------------------------------------------------------------------------------

doleafr:
	iaddiu		ileaftype,vi00,1

createleaf:
	; uses "mtx"
	; [0.0, 0.0,        -radius]		0	0,1
	; [0.0, 0.0,         radius]		1	1,1
	; [0.0, 2.0*radius,  radius]		3	0,0
	; [0.0, 2.0*radius, -radius]		2	1,0
	
	lq.y		leafsize,SETTINGS+0(vi00)
	move.xy		vtx[0],vf00
	move.xy		vtx[1],vf00
	sub.z		vtx[0],vf00,leafsize[y]
	add.z		vtx[1],vf00,leafsize[y]
	move.xz		vtx[2],vtx[1]
	move.xz		vtx[3],vtx[0]
	add.y		vtx[2],leafsize,leafsize[y]
	add.y		vtx[3],leafsize,leafsize[y]

	MatrixLoad		trans,TRANSMTX,vi00
	MatrixLoad		perspmtx,PERSPMTX,vi00

	MatrixMultiply	mtxtmp,mtx,trans				; trans x mtx
	MatrixMultiply	bigmtx,mtxtmp,perspmtx			; Actually (P x T)!!

	MatrixMultiplyVertexW1 vtx[0],bigmtx,vtx[0]
	MatrixMultiplyVertexW1 vtx[1],bigmtx,vtx[1]
	MatrixMultiplyVertexW1 vtx[2],bigmtx,vtx[2]
	MatrixMultiplyVertexW1 vtx[3],bigmtx,vtx[3]

	iaddiu		primdata,vi00,PRIMBUFFER
	ibeq		primbanknum,vi00,rleafprimbankp0
	iaddiu		primdata,vi00,PRIMBUFFER+PRIMSIZEBANK
rleafprimbankp0:

	lq			leafgif,SETTINGS+14(vi00)
	sq			leafgif,0(primdata)

	div			q,vf00[w],vtx[0][w]
	mul.xyz		vtx[0],vtx[0],q
	move		stq[0],vf00
	addq.yz		stq[0],vf00,q
	mr32.w		stq[0],vf00

	div			q,vf00[w],vtx[1][w]
	mul.xyz		vtx[1],vtx[1],q
	move		stq[1],vf00
	addq.xyz	stq[1],vf00,q
	mr32.w		stq[1],vf00

	div			q,vf00[w],vtx[2][w]
	mul.xyz		vtx[2],vtx[2],q
	move		stq[2],vf00
	addq.xz		stq[2],vf00,q
	mr32.w		stq[2],vf00

	div			q,vf00[w],vtx[3][w]
	mul.xyz		vtx[3],vtx[3],q
	move		stq[3],vf00
	addq.z		stq[3],vf00,q
	mr32.w		stq[3],vf00

	loi			0x3FC00000
	addi.x		colsub,vf00,i
	lq			col,COLOURS+0(vi00)
	rnext.x		colrnd,r
	sub.x		colrnd,colrnd,colsub
	loi			0x40800000
	addi.x		colsub,vf00,i
	mul.x		colrnd,colrnd,colsub[x]
	add.y		col,col,colrnd[x]

	mr32.w		vtx[0],vf00
	mr32.w		vtx[1],vf00
	mr32.w		vtx[2],vf00
	mr32.w		vtx[3],vf00

	ftoi4.xyz	vtx[0],vtx[0]
	ftoi4.xyz	vtx[1],vtx[1]
	ftoi4.xyz	vtx[2],vtx[2]
	ftoi4.xyz	vtx[3],vtx[3]

	sq			stq[0],1(primdata)
	sq			col,2(primdata)
	sq			vtx[0],3(primdata)

	sq			stq[1],4(primdata)
	sq			col,5(primdata)
	sq			vtx[1],6(primdata)

	sq			stq[3],7(primdata)
	sq			col,8(primdata)
	sq			vtx[3],9(primdata)

	sq			stq[2],10(primdata)
	sq			col,11(primdata)
	sq			vtx[2],12(primdata)

	xgkick		primdata						; Draw leaf
	iaddiu		flip,vi00,1
	isub		primbanknum,flip,primbanknum	; toggle storage bank
	ibeq		ileaftype,vi00,leftdone

	;------------

rightdone:
	isubiu		level,level,1						; calc parent of current branch
	ibltz		level,alldone

	; pop stack and return
	iaddiu		stackptr,stackptr,STACKFRAMESIZE
	ilw.z		returnaddr,4(stackptr)
	ibeq		returnaddr,vi00,leftdone			; return
	b			rightdone							; return

alldone:

--exit
--endexit

.end

